;
; read.asm - BLKREAD kernal extension handler
;
; Written by
;  Kajtar Zsolt <soci@c64.rulez.org>
;
;    This file is part of IDEDOS the IDE64 disk operating system
;    See README for copyright notice.
;
;    This program is free software; you can redistribute it and/or modify
;    it under the terms of the GNU General Public License as published by
;    the Free Software Foundation; either version 2 of the License, or
;    (at your option) any later version.
;
;    This program is distributed in the hope that it will be useful,
;    but WITHOUT ANY WARRANTY; without even the implied warranty of
;    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
;    GNU General Public License for more details.
;
;    You should have received a copy of the GNU General Public License
;    along with this program; if not, write to the Free Software
;    Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
;
;**************************************************
; Read (A=start address on zeropage, X,Y number of bytes to read)
;
; Used registers: A X Y (X,Y=Number of bytes read)
; Status: C
; C=1:
; A=5 device not present
; A=6 not input file
; A=9 illegal device number
; Status response:
; ST=$00 ok
; ST=$40 end of file
; ST=$42 truncated
; ST=$52 readerror
; ST=$80 device not present
;**************************************************
;----------------------------------------
;IDE read
;
newread		.proc
		jsr read_write_st	;store number of bytes and start address and save zeropage
		#ldatax buffer_input_exp
		jsr read_write_st2	;store current puffer and drive and check state
		lsr
		bcc read_exerr
		lda buffer_type_exp,x
		cmp #BUFFERMODE_FILE
		beq read_file
		cmp #BUFFERMODE_ISOFILE
		beq read_file
		cmp #BUFFERMODE_OFSFILE
		beq read_exerr
		cmp #BUFFERMODE_PCLINK
		beq read_exerr
		lda #$40		;not input file
read_exerr	pha
		jsr OP_restore_zeropage
		pla
		beq ub
		eor #$40
		bne exit_illegal_device
                geq exit_not_input_file

ub		lda #$42
commonend      	sta Status
		ldx #0
                .if SUPERCPU=1
                txy
                .else
		ldy #0
                .fi
		clc
joo		jmp exit2basic

read_file	ldy #0
		tsx
fg		cpx #$03+44		;enough stack space?
		bcs ook
		pla
		sta sp_exp,x		;no, let's backup
		iny
		inx
		gne fg
ook		#phy_a

		#OP_scpu01up
		jsr XIT_read_file

read_write_end	.proc
		.if SUPERCPU=1
		xba			;remember exit status
                #OP_scpu01down
		.else
		sta pom1
		.fi

		#ply_a
		beq kb
		ldx #$03+44
nk		dex
		lda sp_exp,x		;restore stack
		pha
		dey
		bne nk
kb
		.if SUPERCPU=1
		xba
		.else
		lda pom1
		.fi
		bcc *+4
		lda #$52		;on error
		sta Status

		php

		lda read_all_exp
		sec
		sbc read_size
		pha
		lda read_all_exp+1
		sbc read_size+1
		tay

		jsr OP_restore_zeropage

                #plx_a

		plp
                bcc newread.joo
		gcs exit_file_data
		.pend
;----------------------------------------
;file read
;
;BLK_RW_DIRTY, BLK_READ, BLK_RW

XIT_read_file	.proc
bytes_in_buffer	= current_path

		#xit_init

		#ldatax buffer_currentp_exp
		lda buffer_info_exp,x
		pha
		lda #INFO_IGNORE	;be safe
		sta buffer_info_exp,x

		#ldy_l_x_a buffer_address_end_hi_exp,x
		beq middle		;no need to sweep out buffer first

		ldx #read_wr_one_re-read_wr_one_r
le		lda read_wr_one_r-1,x
		sta start_zeropage-1,x	;install zeropage stuff
		dex
		bne le
		lda pom9		;destination
		sta read_wr_one_r.at2+1
		lda pom10
		sta read_wr_one_r.at2+2

		#ldatax buffer_currentp_exp
                sec
                #lda0
		adc buffer_address_low_exp,x
		sta read_wr_one_r.at1+1
		lda buffer_address_hi_exp,x
		and #$7f
		adc #0
		sta read_wr_one_r.at1+2	;Position

		tya
		eor buffer_address_max_hi_exp,x
		cmp #1

		lda buffer_address_end_low_exp,x
		sbc buffer_address_low_exp,x;How many bytes left in buffer?
		sta bytes_in_buffer
		tya
		sbc buffer_address_hi_exp,x
		and #$7f
		sta bytes_in_buffer+1

		ldy read_size
		cpy bytes_in_buffer
		#ldatax read_size+1
		sbc bytes_in_buffer+1
		bcs cf
		sty bytes_in_buffer
		stx bytes_in_buffer+1
cf		php			;remember that we requested less bytes then in buffer

		ldx #0

		ldy bytes_in_buffer+1
		beq ty

ug		jsr start_zeropage	;do copy of 256 bytes
		inc read_wr_one_r.at1+2
		inc read_wr_one_r.at2+2
		dey
		bne ug

ty		ldy bytes_in_buffer
		beq nul

		jsr read_write_part

nul		txa
		clc
		adc read_wr_one_r.at2+1
		sta pom9
		#lda0
		adc read_wr_one_r.at2+2
		sta pom10       	;adjust destination address

		jsr read_write_adj

		plp
		and #$80
		bpl nlst
		lda buffer_address_low_exp,x
		eor buffer_address_end_low_exp,x
		bne nlst		;we read all bytes from buffer

		lda read_size
		ora read_size+1
		beq vw

		ldy #$42		;truncate
vsr		pla
		lda #INFO_BLK_R_EOF
vfe		sta buffer_info_exp,x
		tya
		clc
		rts

vw		ldy #$40		;last byte read
		lda read_all_exp
		ora read_all_exp+1
		bne vsr			;eof
		pla
		#jmps vfe			;eof, but last byte still readable

nlst		ldy #0
		bcc vegeok		;we read less bytes then in buffer, so we are ready

		tya
		sta buffer_address_end_hi_exp,x
		sta buffer_address_end_low_exp,x
		lda buffer_start,x	;mark it as unloaded sector
		#sbc1
		sta buffer_address_hi_exp,x;buffer_address_low_exp is already at 255

		lda buffer_address_max_hi_exp,x
		sbc buffer_start,x
		clc
		adc buffer_filepos2_exp,x
		sta buffer_filepos2_exp,x
		bcc middlex
                jsr incfileposhi
		#jmps middlex
;
;	Now we have to load full sectors to (read_lot_cfs_r.at4), buffer is empty
;
middle		#stz_a bytes_in_buffer
		#stz_aa bytes_in_buffer+1

middlex		lda buffer_filesize1_exp,x
		sta pom1
		lda buffer_filesize2_exp,x
		sec
		sbc buffer_filepos2_exp,x;How much bytes till end of file?
		sta pom2
		lda buffer_filesize3_exp,x
		sbc buffer_filepos3_exp,x
		bne enough
		lda buffer_filesize4_exp,x
		sbc buffer_filepos4_exp,x
		bne enough

		lda read_size
		cmp pom1
		lda read_size+1
		sbc pom2
		bcc enough
		bne vy
		lda pom1
		eor read_size
vy		php
		lda pom2
		sta read_size+1
		lda pom1		;file shorter then expected
		sta read_size
		clc
		adc bytes_in_buffer
		sta read_all_exp
		lda pom2
		adc bytes_in_buffer+1
		sta read_all_exp+1
		.byte SKIP2
enough		clc
		php

		lda read_size
		ora read_size+1
		beq vege2

		#ldatax buffer_currentp_exp
                lda buffer_type_exp,x
                cmp #BUFFERMODE_ISOFILE
                bne cfs

		#psr back2+1
		ldx #load_lot_iso_re-load_lot_iso_r-1
dt2		lda load_lot_iso_r,x
		pha
		dex
		bpl dt2

		tsx
		lda pom9
		sta load_lot_iso_r.at1+1,x
		sta load_lot_iso_r.at2+1,x
		lda pom10
		sta load_lot_iso_r.at1+2,x
		sta load_lot_iso_r.at2+2,x

tnt3		jsr cd_limit_read2047
                pha

		#ldatax buffer_currentp_exp
		lda buffer_filepos2_exp,x
                sta next_sector+3
		lda buffer_filepos3_exp,x
                sta next_sector+2
		lda buffer_filepos4_exp,x

		ldy #3      
ry              lsr 
                ror next_sector+2
                ror next_sector+3
                dey
                bne ry
                tay
                lda next_sector+3
                clc
                adc buffer_isofile_start1,x
                sta next_sector+3
                lda next_sector+2
                adc buffer_isofile_start2,x
                sta next_sector+2
                tya
                adc buffer_isofile_start3,x
                sta next_sector+1
		#lda0
                adc buffer_isofile_start4,x
                sta next_sector

		jsr XIT_upload_next_sector;-=> error

		#ply_a
		jsr OP_read
		bcs handle_error_

		#ldatax buffer_currentp_exp
		lda status_g
		lda low_reg
read_loop2	and #%00001000
		beq handle_error2_	;DRQ=1?

		lda read_size+1
		sec
		sbc #8
		bcc lessthen2048
		sta read_size+1
loo
	   	.if SUPERCPU=1
                ldy #1
                tsx
                phy
                dey
                phx
                .else
		ldy #0
		tsx
		lda #1
		pha
                txa
                pha
                .fi
		rts

handle_error_	#jmps handle_error
handle_error2_	#jmps handle_error2

back2		inc load_lot_iso_r.at1+2,x
		inc load_lot_iso_r.at2+2,x

		#ldatax buffer_currentp_exp
                .if SUPERCPU=1
                lda buffer_filepos2_exp,x
                inc a
                sta buffer_filepos2_exp,x
                .else
                inc buffer_filepos2_exp,x
                lda buffer_filepos2_exp,x
                .fi
                bne fw2
                jsr incfileposhi
fw2

		and #7
                bne loo

		lda status_g
		lda low_reg
		bpl okbsy2
		jsr wait_busy
		bcs handle_error_
		#ldatax buffer_currentp_exp
		lda status_g
		lda low_reg
okbsy2
		#dec_l read_sectors_exp
		bne read_loop2

		and #%00001000
		bne handle_error2_	;DRQ=0?

		lda read_size
		ora read_size+1	;eof!!!!!!!!!!!!!!!!!!!!!!!
                bne tnt3
		geq vege
;
; Now have to load the rest to (read_lot_cfs_r.at1)
;
lessthen2048	lda buffer_start,x
		pha
		jsr OP_read2048_here
                #jmps copyrestofbuffer
;-----------------------------
cfs		#psr back+1
		ldx #load_lot_cfs_re-load_lot_cfs_r-1
dt		lda load_lot_cfs_r,x
		pha
		dex
		bpl dt

		tsx
		lda pom9
		sta load_lot_cfs_r.at1+1,x
		sta load_lot_cfs_r.at3+1,x
		clc
		adc #$80
		sta load_lot_cfs_r.at2+1,x;init pointers
		sta load_lot_cfs_r.at4+1,x
		lda pom10
		sta load_lot_cfs_r.at1+2,x
		adc #0
		sta load_lot_cfs_r.at2+2,x
		#adc1
		sta load_lot_cfs_r.at4+2,x
		lda pom10
		#adc1
		sta load_lot_cfs_r.at3+2,x

tnt2		jsr savesize
		jsr XIT_load_tree	;-=> error
                jsr restoresize

		#ldatax buffer_currentp_exp
		lda buffer_filepos2_exp,x
tnt		lsr
		sta read_p_exp

		lda buffer_start,x
		ora #>b_cfs_file_tree_exp
		tay
		jsr linearaccess

		jsr XIT_upload_next_sector;-=> error

		#ldy_l_a read_sectors_exp
		jsr OP_read
		bcs handle_error

		#ldatax buffer_currentp_exp
		lda status_g
		lda low_reg
read_loop	and #%00001000
		beq handle_error2	;DRQ=1?

		lda read_size+1
		sec
		sbc #2
		bcc lessthen512
		sta read_size+1

		.if SUPERCPU=1
                ldy #1
                tsx
                phy
                dey
                phx
                .else
		ldy #0
		tsx
		lda #1
		pha
                txa
                pha
                .fi
		rts

handle_error2	sec
		lda #EUNEXP
handle_error	jmp LWOP_reporterror_read_xit;-=> error

back		lda load_lot_cfs_r.at1+2,x
		adc #2-1
		sta load_lot_cfs_r.at1+2,x
		#adc1
		sta load_lot_cfs_r.at3+2,x
		lda load_lot_cfs_r.at2+2,x
		adc #2
		sta load_lot_cfs_r.at2+2,x
		#adc1
		sta load_lot_cfs_r.at4+2,x

		#ldatax buffer_currentp_exp
		lda buffer_filepos2_exp,x
		adc #2
		sta buffer_filepos2_exp,x
                bcc fw
                jsr incfileposhi
fw
		lda status_g
		lda low_reg
		bpl okbsy
		jsr wait_busy
		bcs handle_error
		#ldatax buffer_currentp_exp
		lda status_g
		lda low_reg
okbsy
		#dec_l read_sectors_exp
		bne read_loop

		and #%00001000
		bne handle_error2	;DRQ=0?

		lda read_size
		ora read_size+1	;eof!!!!!!!!!!!!!!!!!!!!!!!
		beq vege

		lda buffer_filepos2_exp,x
		bne tnt			;end of a tree sector?
		geq tnt2
;
; Now have to load the rest to (read_lot_cfs_r.at1)
;
lessthen512	lda buffer_start,x
		pha
		tax
		jsr read_buffer_cfs_safe
copyrestofbuffer
		jsr ata_pi_nomore_data
		jsr LWOP_reporterror_read_xit;-=> error

		ldx #read_wr_one_re-read_wr_one_r
le2		lda read_wr_one_r-1,x
		sta start_zeropage-1,x	;install zeropage stuff
		dex
		bne le2

		pla
		sta read_wr_one_r.at1+2

		tsx
		lda load_lot_cfs_r.at1+1,x
		sta read_wr_one_r.at2+1
		lda load_lot_cfs_r.at1+2,x
		sta read_wr_one_r.at2+2

		ldx #0

		ldy read_size+1
		beq rl

rl2		jsr start_zeropage
		inc read_wr_one_r.at1+2
		inc read_wr_one_r.at2+2
		dey
		bne rl2

rl		ldy read_size
		beq rl3

		jsr read_write_part
rl3

		ldy read_wr_one_r.at1+2
		txa
		bne *+3
		dey
                .if SUPERCPU=1
                dec a
                .else
		dex
		txa
                .fi
		#ldx_l_ buffer_currentp_exp
		sta buffer_address_low_exp,x

		lda buffer_address_max_hi_exp,x
		sta buffer_address_end_hi_exp,x
		#lda0
		sta buffer_address_end_low_exp,x
		sta read_size
		sta read_size+1

		lda buffer_filesize1_exp,x
		sec
		sbc #1
		sta pom1
		lda buffer_filesize2_exp,x
		sbc buffer_filepos2_exp,x
		sta pom2
		lda buffer_filesize3_exp,x
		sbc buffer_filepos3_exp,x
		bne more
		lda buffer_filesize4_exp,x
		sbc buffer_filepos4_exp,x
		bcc more
		bne more

		lda buffer_address_max_hi_exp,x
		sbc buffer_start,x
		clc
		sbc pom2
		bcc more

		lda buffer_start,x
		ora pom2
		sta buffer_address_end_hi_exp,x
		lda pom1
		sta buffer_address_end_low_exp,x
more
		tya
		cmp buffer_address_end_hi_exp,x
		bne *+4
		ora #$80
		sta buffer_address_hi_exp,x

vege	       
		.if SUPERCPU=1
		lda xit_sp_exp
                dec a
                dec a
		tcs
                .else
		ldx xit_sp_exp
                dex
                dex
		txs
                .fi

vege2		ldy #0
		#ldatax buffer_currentp_exp
		plp
		bcc vegeok
		beq nt
		ldy #$42
		.byte SKIP2
nt		ldy #$40
		pla
		lda #INFO_BLK_R_EOF
		.byte SKIP1
vegeok		pla			;be safe
		sta buffer_info_exp,x

		tya
		clc
		rts			;-=> ok
		.pend
;----------------------------------------
;Common start2 for blkread/write
;
read_write_st2	.proc
		sta buffer_currentp_exp
		lda buffer_drive_exp,x
		sta current_drive_exp
		lda buffer_info_exp,x
		and #INFO_MSK
		lsr
		tay
		lda brw_good,y
		rts

;
;bit 7: 1=writeable
;bit 1: 1=illegal device
;bit 0: 1=readable
;
brw_good	.byte $80		;INFO_BLK_WRITE_DIRTY
		.byte $81		;INFO_BLK_RW_DIRTY
		.byte $80		;INFO_BLK_RW_EOF_DIRTY
		.byte $80		;INFO_BLK_RW_EOF
		.byte $00		;INFO_BLK_RW_EOF_FULL
		.byte $00		;INFO_BLK_WRITE_FULL
		.byte $80		;INFO_BLK_WRITE
		.byte $81		;INFO_BLK_RW
		.byte $01		;INFO_BLK_READ
		.byte $00		;INFO_BLK_R_EOF
		.byte $02		;INFO_STATUS
		.byte $02		;INFO_KERNAL
		.byte $02		;INFO_IGNORE
		.pend
;----------------------------------------
;buffer copyer for blkread
;
read_wr_one_r	.proc
		.logical start_zeropage
		.if SUPERCPU=1
		sta cfg_stnd		;3
at1		lda $01ff00,x		;4
at2		sta $ffff,x		;3
		inx
at3             bne at1
		nop
                jmp switch_16k		;SCPU bug!
		.else
at1		lda $ff00,x		;3
		sta cfg_stnd		;3
at2		sta $ffff,x		;3
                sta cfg_open
                inx
at3		bne at1			;2
		rts			;1
		rts			;1 18 max
		.fi
		.here
		.pend
read_wr_one_re
;----------------------------------------
;Common start for blkread/write
;
read_write_st	.proc
		pha
		#sty_l_a read_all_exp+1
		#stx_l_a read_all_exp

		jsr OP_save_zeropage

		sty read_size+1
                lda read_all_exp
		sta read_size

		#plx_a
		ldy $00,x
		lda $01,x
		sty pom9		;get destination address
		sta pom10
		rts
		.pend
;----------------------------------------
;Less then 256 bytes copy for blkread/write
;
read_write_part	.proc
		lda #$88		;dey
		sta read_wr_one_r.at3
		lda #$d0		;bne
		sta read_wr_one_r.at3+1
		.if SUPERCPU=1
		lda #$fe-1-1-4-3
		.else
		lda #$fe-1-1-3-3-3-3
		.fi
		sta read_wr_one_r.at3+2

		jmp start_zeropage
		.pend
;----------------------------------------
;Remaining bytes and bufferpos adjust for blkread/write
;
read_write_adj	.proc
bytes_in_buffer	= current_path

		lda read_size
		sec
		sbc bytes_in_buffer
		sta read_size	;substract copyend bytes
		lda read_size+1
		sbc bytes_in_buffer+1
		sta read_size+1

		.if SUPERCPU=1
		lda buffer_currentp_exp
		tax
		lda bytes_in_buffer
		.else
		txa
		ldx buffer_currentp_exp
		.fi
		clc
		adc buffer_address_low_exp,x
		sta buffer_address_low_exp,x;adjust buffer position
		lda buffer_address_hi_exp,x
		and #$7f
		adc bytes_in_buffer+1
		cmp buffer_address_end_hi_exp,x
		bne *+4
		ora #$80		;in last bank!
		sta buffer_address_hi_exp,x
		rts
		.pend
;----------------------------------------
;Increment high part of filepos
;
incfileposhi	.proc
		.if SUPERCPU=1
                xba
                lda buffer_filepos3_exp,x
                inc a
                sta buffer_filepos3_exp,x
                bne fw3
                lda buffer_filepos4_exp,x
                inc a
                sta buffer_filepos4_exp,x
fw3             xba
                rts
                .else
		inc buffer_filepos3_exp,x
		bne fw3
		inc buffer_filepos4_exp,x;!!!!!
fw3		rts
		.fi
		.pend
;----------------------------------------
savesize	.proc
		lda load_size
                sta load_size_exp
		lda load_size+1
                sta load_size_exp+1
                rts
                .pend

read_buffer_cfs_safe .proc
		jsr savesize
		jsr read_buffer_cfs

restoresize	.proc
                lda load_size_exp
		sta load_size
                lda load_size_exp+1
		sta load_size+1
                rts
                .pend
